import json

from django.contrib import admin
from django.utils.html import format_html

import boto3
from botocore.exceptions import ClientError

from mbq.pubsub import models


def get_name_from_topic_arn(arn):
    return arn.split(":")[-1]


class SNSTopicListFilter(admin.SimpleListFilter):
    title = "topic"
    parameter_name = "topic_arn"

    def lookups(self, request, model_admin):
        qs = model_admin.get_queryset(request).order_by("topic_arn")
        topic_arns = qs.values_list("topic_arn", flat=True).distinct()
        for arn in topic_arns:
            yield (arn, get_name_from_topic_arn(arn))

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(topic_arn=self.value())
        return queryset


def replay_messages(model_admin, request, queryset):
    sqs = boto3.client("sqs")
    queue_urls = {}

    for message in queryset:
        if message.queue not in queue_urls:
            url = sqs.get_queue_url(QueueName=message.queue)
            queue_urls[message.queue] = url["QueueUrl"]

        try:
            sqs.send_message(QueueUrl=queue_urls[message.queue], MessageBody=message.payload)
            message.delete()
        except ClientError as e:
            model_admin.message_user(
                request, f"Message {message.id} failed to replay: {e.response}"
            )


replay_messages.short_description = "Replay messages"


class UndeliverableMessageAdmin(admin.ModelAdmin):
    list_display = ("id", "created_at", "message_timestamp", "message_type", "queue")

    list_filter = ("message_type", "queue")

    fields = ("id", "created_at", "message_timestamp", "message_type", "queue", "topic_arn")

    readonly_fields = (
        "id",
        "created_at",
        "message_timestamp",
        "message_type",
        "queue",
        "topic_arn",
    )

    def has_delete_permission(self, *args):
        return True

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, *args):
        return False

    def admin_topic(self, message):
        return get_name_from_topic_arn(message.topic_arn)

    admin_topic.short_description = "topic"

    list_display = list_display + ("admin_topic",)
    list_filter = list_filter + (SNSTopicListFilter,)

    def admin_payload(self, message):
        try:
            payload = json.loads(message.payload)
            data = json.loads(payload["Message"])
            payload["Message"] = data
            payload_str = json.dumps(payload, indent=4)
            return format_html("<br/><pre>{}</pre>", payload_str)
        except Exception:
            return format_html("<br/><pre>{}</pre>", message.payload)

    admin_payload.short_description = "payload"

    fields = fields + ("admin_payload",)
    readonly_fields = readonly_fields + ("admin_payload",)

    actions = [replay_messages]


admin.site.register(models.UndeliverableMessage, UndeliverableMessageAdmin)



from contextlib import ContextDecorator
from importlib import import_module

from django.conf import settings
from django.dispatch import receiver
from django.test.signals import setting_changed

# TODO add checking for required settings
DEFAULTS = {
    "ENV": None,
    "SERVICE": None,
    "QUEUES": [],
    "CONSUMER_CLASS": "mbq.pubsub.consumer.Consumer",
    "MESSAGE_HANDLERS": {},
    "USE_DATABASE": True,
}

IMPORT_STRINGS = {"CONSUMER_CLASS", "MESSAGE_HANDLERS"}


# The following is heavily inspired by the approach used in
# Django Restframework: https://github.com/encode
# /django-rest-framework/blob/master/rest_framework/settings.py
#
# Copyright © 2011-present, Encode OSS Ltd. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


def perform_import(val, setting_name):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if val is None:
        return None
    elif isinstance(val, str):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    elif isinstance(val, dict):
        return {k: import_from_string(v, setting_name) for k, v in val.items()}
    return val


def import_from_string(val, setting_name):
    """
    Attempt to import a class from a string representation.
    """
    try:
        # Nod to tastypie's use of importlib.
        module_path, class_name = val.rsplit(".", 1)
        module = import_module(module_path)
        return getattr(module, class_name)
    except (ImportError, AttributeError) as e:
        raise ImportError(
            f"Could not import '{val}' for PubSub setting '{setting_name}'. {e.__class__.__name__}: {e}."
        )


class ProjectSettings:
    def __init__(self, user_settings=None, defaults=None, import_strings=None):
        if user_settings:
            self._user_settings = self.__check_user_settings(user_settings)
        self.defaults = defaults or DEFAULTS
        self.import_strings = import_strings or IMPORT_STRINGS
        self._cached_attrs = set()

    @property
    def user_settings(self):
        if not hasattr(self, "_user_settings"):
            self._user_settings = getattr(settings, "PUBSUB", {})
        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid PUBSUB setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Coerce import strings into classes
        if attr in self.import_strings:
            val = perform_import(val, attr)

        # Cache the result
        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, "_user_settings"):
            delattr(self, "_user_settings")

    def __call__(self, **overrides):
        """
        ContextDecorator for use in testing.
        """

        settings = self
        originals = {}
        UNSET = object()

        class contextdecorator(ContextDecorator):
            def __enter__(self):
                for k, v in overrides.items():
                    if hasattr(settings, k):
                        originals[k] = getattr(settings, k)
                    else:
                        originals[k] = UNSET
                    setattr(settings, k, v)

            def __exit__(self, *exc):
                for k, v in originals.items():
                    if v is UNSET:
                        delattr(settings, k)
                    else:
                        setattr(settings, k, v)

        return contextdecorator()


project_settings = ProjectSettings(None, DEFAULTS, IMPORT_STRINGS)


@receiver(setting_changed)
def reload_api_settings(*args, **kwargs):
    setting = kwargs["setting"]
    if setting == "PUBSUB":
        project_settings.reload()



__author__ = "Managed by Q, Inc."
__author_email__ = "open-source@managedbyq.com"
__description__ = "MBQ PubSub"
__license__ = "Apache 2.0"
__title__ = "mbq.pubsub"
__url__ = "https://github.com/managedbyq/mbq.pubsub"
__version__ = "0.0.3"



import json
import logging

import arrow
import boto3
import rollbar

from . import _collector as collector
from . import exceptions, models
from .settings import project_settings

logger = logging.getLogger(__name__)

NOT_PROVIDED = object()


class Consumer:
    def __init__(self, queue_name: str, handlers: dict):
        self._queue_name = queue_name
        self._queue_full_name = (
            f"mbq-{project_settings.SERVICE}-{queue_name}-{project_settings.ENV.short_name}"
        )
        self._handlers = handlers

    @property
    def queue(self):
        if not hasattr(self, "_queue"):
            sqs = boto3.resource("sqs")
            self._queue = sqs.get_queue_by_name(QueueName=self._queue_full_name)
        return self._queue

    @property
    def dead_letter_queue(self):
        """
        Find the dead letter queue from the primary queue's redrive policy.

        https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-dead-letter-queues.html
        """

        if not hasattr(self, "_dead_letter_queue"):
            try:
                redrive_policy = json.loads(self.queue.attributes["RedrivePolicy"])
                dlq_name = redrive_policy["deadLetterTargetArn"].split(":")[-1]
            except Exception as e:
                raise exceptions.ConsumerException(
                    f"No dead letter queue configured for {self.queue}"
                ) from e

            sqs = boto3.resource("sqs")
            self._dead_letter_queue = sqs.get_queue_by_name(QueueName=dlq_name)
        return self._dead_letter_queue

    def process_queue(self):
        collector.increment("consumer.attempt_read", tags={"queue": self._queue_name})
        messages = self.queue.receive_messages(WaitTimeSeconds=5, MaxNumberOfMessages=10)
        if len(messages) > 0:
            logger.info(f"Received {len(messages)} messages")

        for message in messages:
            try:
                body = json.loads(message.body)
                data = json.loads(body["Message"])
                message_type = data["message_type"]
                payload = data["payload"]
            except Exception:
                rollbar.report_exc_info()
                continue
            try:
                handler = self._handlers.get(message_type)
                if handler:
                    logger.info(f"Processing {message_type} message")
                    handler(payload)
                    result = "succeeded"
                    logger.info(f"Done processing {message_type} message")

                else:
                    result = "skipped"
                    logger.info(
                        f"Received unregistered message_type: {message_type} \n"
                        f"Message body: {body}"
                    )
            except Exception:
                result = "failed"
                logger.exception("An error occurred while processing the message.")
                rollbar.report_exc_info()
            else:
                message.delete()

            collector.increment(
                "consumer.processed",
                tags={"result": result, "message_type": message_type, "queue": self._queue_name},
            )

    def process_dead_letter_queue(self):
        messages = self.dead_letter_queue.receive_messages(
            WaitTimeSeconds=0, MaxNumberOfMessages=10
        )
        if len(messages) > 0:
            logger.info(f"Received {len(messages)} messages on the dead letter queue")

        for message in messages:
            body = json.loads(message.body)
            data = json.loads(body["Message"])
            models.UndeliverableMessage.objects.create(
                message_type=data["message_type"],
                message_timestamp=arrow.get(body["Timestamp"]).datetime,
                payload=message.body,
                queue=self._queue_full_name,
                topic_arn=body.get("TopicArn"),
            )
            message.delete()

    def replay_dead_letter_queue(self, max_messages: int):
        if max_messages < 1:
            raise exceptions.ConsumerException("max_messages must be greater than 0")

        logger.info(
            f"Replaying at most {max_messages} messages "
            f"from {self.dead_letter_queue} to {self.queue}"
        )

        messages_processed = 0
        while True:
            messages = self.dead_letter_queue.receive_messages(
                WaitTimeSeconds=20, MaxNumberOfMessages=10
            )

            for message in messages:
                self.queue.send_message(MessageBody=message.body)
                message.delete()
                messages_processed += 1

                if messages_processed >= max_messages:
                    return



# Generated by Django 2.1.5 on 2019-01-11 22:34

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="UndeliverableMessage",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, editable=False, primary_key=True, serialize=False
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("message_type", models.CharField(max_length=256)),
                ("message_timestamp", models.DateTimeField()),
                ("payload", models.TextField()),
                ("queue", models.CharField(max_length=80)),
                ("topic_arn", models.CharField(max_length=2048, null=True)),
            ],
        )
    ]



import os

from mbq import env
import boto3
import dj_database_url


BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

ALLOWED_HOSTS = ["pubsub.lcl.mbq.io"]

DEBUG = True

DEBUG_PROPAGATE_EXCEPTIONS = True

DATABASE_URL = env.get("DATABASE_URL", required=False)

if DATABASE_URL is not None:
    DATABASES = {
        "default": dj_database_url.parse(
            DATABASE_URL, engine="django.db.backends.postgresql", conn_max_age=0
        )
    }
else:
    DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}}

SITE_ID = 1

SECRET_KEY = "BACON"

USE_I18N = True

USE_L10N = True

STATIC_URL = "/static/"

STATIC_ROOT = os.path.join(BASE_DIR, "staticfiles")


ROOT_URLCONF = "tests.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "debug": True,
            "context_processors": {
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            },
        },
    }
]

MIDDLEWARE = [
    "django.middleware.common.CommonMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
]

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.staticfiles",
    "django.contrib.messages",
    "mbq.pubsub",
    "tests",
]

PASSWORD_HASHERS = ["django.contrib.auth.hashers.MD5PasswordHasher"]

USE_TZ = True

MESSAGE_HANDLERS = "tests.message_handlers"

PUBSUB = {
    "ENV": env.Environment.LOCAL,
    "SERVICE": "pubsub",
    "QUEUES": ["updates"],
    "MESSAGE_HANDLERS": {
        "pubsub.update_raise_exception": f"{MESSAGE_HANDLERS}.handle_update_raise_exception"
    },
}

LOGGING_LEVEL = "DEBUG"

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "verbose": {
            "format": (
                "%(levelname)-8s %(asctime)s %(name)s " "%(filename)s:%(lineno)s %(message)s"
            )
        }
    },
    "handlers": {
        "console": {"level": "INFO", "class": "logging.StreamHandler", "formatter": "verbose"}
    },
    "loggers": {
        "django.db.backends": {"handlers": ["console"], "level": LOGGING_LEVEL, "propagate": False},
        "botocore": {"handlers": ["console"], "level": "ERROR", "propagate": False},
        "boto3": {"handlers": ["console"], "level": "ERROR", "propagate": False},
        "": {"handlers": ["console"], "level": LOGGING_LEVEL},
    },
}

AWS_ACCESS_KEY_ID = env.get("AWS_ACCESS_KEY_ID", required=False)
AWS_DEFAULT_REGION = env.get("AWS_DEFAULT_REGION", required=False)
AWS_SECRET_ACCESS_KEY = env.get("AWS_SECRET_ACCESS_KEY", required=False)

if AWS_ACCESS_KEY_ID is not None:
    boto3.setup_default_session(
        aws_access_key_id=AWS_ACCESS_KEY_ID,
        aws_secret_access_key=AWS_SECRET_ACCESS_KEY,
        region_name=AWS_DEFAULT_REGION,
    )



from django.conf import settings
from django.conf.urls import url
from django.conf.urls.static import static
from django.contrib import admin

urlpatterns = [url(r"^admin/", admin.site.urls)]

urlpatterns += static(settings.STATIC_URL, document_root=settings.STATIC_ROOT)
